#include <qobject.h>
#include <qmutex.h>
#include <qstring.h>
#include <qtimer.h>

#include "Account.hpp"
#include "Session.hpp"

class Call;

class PhoneLine : public QObject
{
  Q_OBJECT
  
public:
  PhoneLine(const Session &session, 
	    const Account &account, 
	    unsigned int line);
  ~PhoneLine();

  void call(const QString &to);
  void call();
  void answer();
  void hangup(bool sendrequest = true);
  void hold();
  void unhold();

  QString getCallId();

  unsigned int line();

  /**
   * This will lock the current phone line.
   * 
   * Note: this will only lock the phone line
   * for those that uses this lock, unlock
   * mechanism. PhoneLineManager always uses
   * this mechanism. So, if you work only with
   * PhoneLineManager, it will be thread safe.
   */
  void lock();

  /**
   * This will unlock the current phone line.
   * See the Note of the lock function.
   */
  void unlock();


  /**
   * This function will return true if there's no 
   * activity on this line. It means that even 
   * if we typed something on this line, but haven't
   * started any communication, this will be available.
   */
  bool isAvailable()
  {return !mCall;}

  void sendKey(Qt::Key c);

  QString getLineStatus();
  QString getBuffer()
  {return mBuffer;}
  
public slots:
  void setLineStatus(const QString &);
  void setAction(const QString &);
  void setTempAction(const QString &);
  void resetAction();
  void incomming(const Call &call);

  /**
   * Clears the buffer of the line.
   */
  void clear();
  
  /**
   * The user selected this line.
   */
  void select(bool hardselect = false);

  /**
   * This phoneline is no longer selected.
   */
  void unselect(bool hardselect = false);

  /**
   * This will do a hard unselect. it means it
   * will remove the call if there's one.
   */
  void disconnect();

  /**
   * This will close the current call. it means it
   * will remove the call if there's one.
   */
  void close();

  /**
   * This will close the current call. it means it
   * will remove the call if there's one. The line
   * will be in an error state.
   */
  void error();

  void setState(const QString &){}
  void setPeer(const QString &){}


signals:
  void selected();
  void unselected();
  void backgrounded();
  void lineStatusChanged(QString);
  void actionChanged(QString);
  void bufferStatusChanged(QString);

private:
  Session mSession;
  Account mAccount;
  Call *mCall;
  QMutex mPhoneLineMutex;
  unsigned int mLine;

  bool mSelected;
  bool mInUse;
  //This is the buffer when the line is not in use;
  QString mBuffer;

  QString mLineStatus;
  QString mAction;
  QTimer *mActionTimer;

  bool mIsOnError;
};
